##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################


from http import HTTPStatus

from flask import Blueprint, Response, make_response, request
from marshmallow import EXCLUDE, Schema, post_load
from marshmallow.fields import Email, Str

from f4ratk.web.infrastructure import di
from f4ratk.web.mail.mail import MailRequest, MailService

mail_blueprint = Blueprint(name='mail', import_name=__name__)


class MailRequestSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = Str()
    contact = Email()
    content = Str()

    address = Str()
    additionalName = Str()

    @post_load
    def to_mail_request(self, data: dict, **_):
        bot = (data.get('address') or data.get('additionalName')) is not None

        return MailRequest(
            name=data.get('name'),
            contact=data.get('contact'),
            content=data.get('content'),
            bot=bot,
        )


@mail_blueprint.route('/v0/mails', methods=['POST'])
def mails() -> Response:
    mail_request: MailRequest = MailRequestSchema().load(request.get_json())

    di[MailService].relay(mail_request)

    return make_response(
        '', HTTPStatus.NO_CONTENT, {'Content-Type': 'application/json'}
    )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

from pandas import DataFrame

from f4ratk.analyze.evaluation import EvaluatedResults
from f4ratk.domain import AnalysisConfig, Currency, Frame


@dataclass(frozen=True)
class Stock:
    ticker_symbol: str
    currency: Currency = Currency.USD


class NoTickerData(IOError):
    def __init__(self, symbol: str):
        super().__init__(f"No data fetched for symbol '{symbol}'.")


class TickerAnalyzer(metaclass=ABCMeta):
    @abstractmethod
    def analyze_ticker_symbol(
        self, stock: Stock, analysis_config: AnalysisConfig
    ) -> EvaluatedResults:
        """Analyze ticker symbol data."""


class TickerReader(metaclass=ABCMeta):
    @abstractmethod
    def read_ticker(self, stock: Stock, frame: Frame) -> DataFrame:
        """Read ticker symbol data to DataFrame."""



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from typing import Callable, Tuple, Union
from unittest.mock import Mock

from pandas import DataFrame, DatetimeIndex, PeriodIndex
from pandas_datareader.yahoo.daily import YahooDailyReader
from pytest import fixture
from pytest_mock import MockerFixture

from f4ratk.fama import FamaReader

Quote = Return = Tuple[str, float]
QuotesFactory = FamaFactory = ReturnsFactory = Callable[[Tuple[Return, ...]], DataFrame]

YahooReaderFactory = Callable[[Tuple[Quote, ...]], YahooDailyReader]
FamaReaderFactory = Callable[[Tuple[Return, ...]], FamaReader]


@fixture(scope='module', name="create_fama")
def fama_factory() -> FamaFactory:
    def _fama_factory(*returns: Return):
        periods, returns = list(zip(*returns))
        data = DataFrame(
            data={
                'MKT': returns,
                'SMB': returns,
                'HML': returns,
                'RMW': returns,
                'CMA': returns,
                'WML': returns,
                'RF': returns,
            },
            index=PeriodIndex(data=periods, dtype='period[M]'),
        )

        return data

    return _fama_factory


@fixture(scope='module', name="create_quotes")
def quotes_factory() -> QuotesFactory:
    def _quotes_factory(
        *quotes: Quote, index: Union[PeriodIndex, DatetimeIndex] = PeriodIndex
    ):
        index_data, returns = list(zip(*quotes))

        def create_index() -> Union[PeriodIndex, DatetimeIndex]:
            if index is PeriodIndex:
                return PeriodIndex(data=index_data, dtype='period[M]')
            elif index is DatetimeIndex:
                return DatetimeIndex(data=index_data, dtype='datetime64[ns]')

        data = DataFrame(
            data={'Adj Close': returns},
            index=create_index(),
        )

        return data

    return _quotes_factory


@fixture(scope='module', name="create_returns")
def returns_factory() -> ReturnsFactory:
    def _returns_factory(
        *quotes: Quote, index: Union[PeriodIndex, DatetimeIndex] = PeriodIndex
    ):
        index_data, returns = list(zip(*quotes))

        def create_index() -> Union[PeriodIndex, DatetimeIndex]:
            if index is PeriodIndex:
                return PeriodIndex(data=index_data, dtype='period[M]')
            elif index is DatetimeIndex:
                return DatetimeIndex(data=index_data, dtype='datetime64[ns]')

        data = DataFrame(
            data={'Returns': returns},
            index=create_index(),
        )

        return data

    return _returns_factory


@fixture(scope='function', name="yahoo_returns")
def yahoo_reader_factory(
    mocker: MockerFixture, create_quotes: QuotesFactory
) -> YahooReaderFactory:
    def _yahoo_reader_factory(*quotes: Quote):
        mock_reader = Mock()
        mock_reader.read.return_value = create_quotes(
            *quotes,
            index=DatetimeIndex,
        )
        mocker.patch(
            'f4ratk.ticker.reader.yahoo_reader',
            spec_set=YahooDailyReader,
            return_value=mock_reader,
        )

    return _yahoo_reader_factory


@fixture(scope='function', name="fama_returns")
def fama_reader_factory(
    mocker: MockerFixture, create_fama: FamaFactory
) -> FamaReaderFactory:
    def _fama_reader_factory(*returns: Return):
        returns = create_fama(*returns)
        mocker.patch(
            'f4ratk.infrastructure.FamaReader',
            spec_set=FamaReader,
            **{'fama_data.return_value': returns},
        )

    return _fama_reader_factory



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import ssl
from dataclasses import dataclass
from email.message import EmailMessage
from smtplib import SMTP
from typing import Optional


@dataclass(frozen=True)
class MailRequest:
    name: Optional[str]
    contact: Optional[str]
    content: Optional[str]
    bot: bool


@dataclass(frozen=True)
class MailConfig:
    from_address: str
    to_address: str


class MailComposer:
    def __init__(self, mail_config: MailConfig):
        self._mail_config = mail_config

    def mail(self, request: MailRequest) -> EmailMessage:
        sender = self._format_sender(request)
        bot = "[BOT]" if request.bot else ""

        message = EmailMessage()
        if request.content:
            message.set_content(request.content)
        message['Subject'] = f"[F4RATK]{bot} Feedback from '{sender}'"
        message['From'] = self._mail_config.from_address
        message['To'] = self._mail_config.to_address

        return message

    @staticmethod
    def _format_sender(request: MailRequest) -> str:
        if request.name and request.contact:
            return f"{request.name} ({request.contact})"
        elif request.name or request.contact:
            return request.name or request.contact
        else:
            return "Anonymous User"


@dataclass(frozen=True)
class ServerConfig:
    username: str
    password: str
    host: str
    port: int


class MailMessageService:
    def __init__(self, server_config: ServerConfig):
        self._server_config = server_config

    def send(self, message: EmailMessage):
        with self._server_connection() as server:
            server.starttls(context=ssl.create_default_context())
            server.login(
                user=self._server_config.username, password=self._server_config.password
            )
            server.send_message(message)
            server.quit()

    def _server_connection(self) -> SMTP:
        return SMTP(
            host=self._server_config.host,
            port=self._server_config.port,
        )


class MailService:
    def __init__(self, composer: MailComposer, messenger: MailMessageService):
        self._composer = composer
        self._messenger = messenger

    def relay(self, request: MailRequest):
        self._messenger.send(self._composer.mail(request))



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import Enum, unique
from os import PathLike

from pandas import DataFrame

from f4ratk.analyze.evaluation import EvaluatedResults
from f4ratk.domain import AnalysisConfig, Currency, Frame


@unique
class ValueFormat(Enum):
    PRICE = 'Prices'
    RETURN = 'Returns'


@dataclass(frozen=True)
class FileConfig:
    path: PathLike
    currency: Currency
    value_format: ValueFormat


class FileAnalyzer(metaclass=ABCMeta):
    @abstractmethod
    def analyze_file(
        self,
        file_config: FileConfig,
        analysis_config: AnalysisConfig,
    ) -> EvaluatedResults:
        """Analyze file content."""


class FileReader(metaclass=ABCMeta):
    @abstractmethod
    def read(self, file_config: FileConfig, frame: Frame) -> DataFrame:
        """Read file content to DataFrame."""



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import logging
from os import environ, getenv
from pathlib import Path
from typing import Any, Dict, Optional, Type, Union

from flask_caching import Cache

from f4ratk.directories import cache
from f4ratk.web.mail.mail import (
    MailComposer,
    MailConfig,
    MailMessageService,
    MailService,
    ServerConfig,
)

Dependencies = Dict[Type[Any], Any]

di: Dependencies = dict()

log = logging.getLogger(__name__)


def locate_cache_dir() -> Union[str, Path]:
    return getenv('CACHE_DIR', cache.file('server'))


def create_cache() -> Cache:
    cache_type = getenv('CACHE_TYPE')
    cache_dir = locate_cache_dir() if cache_type == 'FileSystemCache' else None

    config = dict()

    if cache_type:
        config['CACHE_TYPE'] = cache_type

    if cache_dir:
        config['CACHE_DIR'] = cache_dir

    return Cache(config=config)


server_cache = create_cache()


def mail_config() -> MailConfig:
    return MailConfig(
        from_address=environ['MAIL_FROM_ADDRESS'], to_address=environ['MAIL_TO_ADDRESS']
    )


def server_config() -> ServerConfig:
    return ServerConfig(
        username=environ['MAIL_SERVER_USERNAME'],
        password=environ['MAIL_SERVER_PASSWORD'],
        host=environ['MAIL_SERVER_HOST'],
        port=environ.get('MAIL_SERVER_PORT', 587),
    )


def mail_service() -> Optional[MailService]:
    try:
        return MailService(
            composer=MailComposer(mail_config=mail_config()),
            messenger=MailMessageService(server_config=server_config()),
        )
    except KeyError as error:
        log.warning(
            f"Deactivating the mail feature due to incomplete configuration (Cause: {error})."  # noqa: E501
        )
        return None


def instantiate_dependencies() -> None:
    log.debug("Bootstrapping mail dependencies")
    di.update({MailService: mail_service()})


def dependencies() -> Dict[Type[Any], Any]:
    return di


def mail_enabled() -> bool:
    return di[MailService] is not None



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import logging
from dataclasses import dataclass
from json import loads
from typing import List, Tuple

from flask import Blueprint, Response, after_this_request, jsonify, request
from pandas import DataFrame, Series, concat
from statsmodels.regression.linear_model import RegressionResultsWrapper

# noinspection PyPackageRequirements
from werkzeug.datastructures import FileStorage

from f4ratk.analyze.evaluation import EvaluatedResult
from f4ratk.cli.types import RegionChoice
from f4ratk.domain import AnalysisConfig, Currency, Frame, Frequency
from f4ratk.file.api import FileAnalyzer
from f4ratk.file.reader import FileConfig, ValueFormat
from f4ratk.infrastructure import di
from f4ratk.shared import first_period, last_period
from f4ratk.ticker.api import TickerAnalyzer
from f4ratk.ticker.reader import Stock
from f4ratk.web.infrastructure import server_cache

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class Coefficient:
    name: str
    weight: float
    standardError: float
    probability: float
    confidenceStart: float
    confidenceEnd: float


@dataclass(frozen=True)
class Period:
    start: str
    end: str


@dataclass(frozen=True)
class AnalysisReport:
    coefficients: Tuple[Coefficient]
    confidenceLevel: float
    adjustedRSquared: float
    observations: int
    excessReturn: float
    period: Period


_DEFAULT_QUERY_CACHE = 60 * 60 * 24 * 14

web_blueprint = Blueprint(name='web', import_name=__name__)


@web_blueprint.route('/v0/tickers/<symbol>;region=<region>', methods=['GET'])
@server_cache.cached(timeout=_DEFAULT_QUERY_CACHE)
def tickers(symbol, region) -> Response:
    stock = Stock(ticker_symbol=symbol, currency=Currency.USD)

    analysis_config = AnalysisConfig(
        region=RegionChoice().convert(region, None, None),
        frame=Frame(frequency=Frequency.MONTHLY, start=None, end=None),
    )

    results = di[TickerAnalyzer].analyze_ticker_symbol(
        stock=stock,
        analysis_config=analysis_config,
    )

    report = to_report(results.ff6)

    @after_this_request
    def add_cache_control(response):
        response.headers['Cache-Control'] = f'max-age={_DEFAULT_QUERY_CACHE}'
        return response

    return jsonify(report)


@web_blueprint.route('/v0/files', methods=['POST'])
def files() -> Response:
    received_files: List[FileStorage] = request.files.getlist(key='file')
    config = loads(request.form.get(key='config', default=None))
    file = next(file for file in received_files)

    results = di[FileAnalyzer].analyze_file(
        FileConfig(file, Currency.USD, ValueFormat.PRICE),
        AnalysisConfig(
            RegionChoice().convert(config['region'], None, None),
            Frame(Frequency.MONTHLY, None, None),
        ),
    )

    report = to_report(results.ff6)

    return jsonify(report)


def to_report(result: EvaluatedResult) -> AnalysisReport:
    confidence_level = 0.95

    return AnalysisReport(
        coefficients=to_coefficients(result.model, confidence_level),
        confidenceLevel=confidence_level,
        adjustedRSquared=result.model.rsquared_adj,
        observations=int(result.model.nobs),
        excessReturn=result.evaluation,
        period=to_period(result.model),
    )


def to_period(model: RegressionResultsWrapper) -> Period:
    residuals: Series = model.resid
    return Period(start=str(first_period(residuals)), end=str(last_period(residuals)))


def to_coefficients(
    model: RegressionResultsWrapper, confidence_level: float
) -> Tuple[Coefficient]:
    alpha = 1 - confidence_level

    joined: DataFrame = concat(
        (model.params, model.bse, model.pvalues, model.conf_int(alpha=alpha)), axis=1
    )

    return tuple(
        Coefficient(name, weight, std, probability, conf_start, conf_end)
        for name, weight, std, probability, conf_start, conf_end in joined.itertuples()
    )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from unittest.mock import Mock

from pandas import Period
from pytest import approx

from f4ratk.domain import Frequency, Region
from f4ratk.fama import FamaReader
from f4ratk.history import Historian
from tests.conftest import FamaFactory


class TestAnnualizedReturns:
    def given_factor_returns_when_displaying_history_for_region_should_calc_annualized_returns_from_monthly_fama_data(  # noqa: E501
        self, create_fama: FamaFactory
    ):
        returns = create_fama(
            ('2020-04', -5.0),
            ('2020-05', 10.0),
            ('2020-06', -5.0),
            ('2020-07', 10.0),
            ('2020-08', -5.0),
            ('2020-09', 10.0),
            ('2020-10', -5.0),
            ('2020-11', 10.0),
            ('2020-12', -5.0),
            ('2021-01', 10.0),
            ('2021-02', -5.0),
            ('2021-03', 10.0),
            ('2021-04', -5.0),
        )

        fama_reader = Mock(spec_set=FamaReader)
        fama_reader.fama_data.return_value = returns

        result = Historian(fama_reader).annualized_returns(region=Region.US)

        assert list(result._data.columns) == ['Returns', 'SD']
        assert result._data['Returns']['MKT'] == approx(21.706, rel=0.0001)
        assert result._data['SD']['MKT'] == approx(26.961, rel=0.0001)

        assert result._first == Period('2020-04', 'M')
        assert result._last == Period('2021-04', 'M')

        fama_reader.fama_data.assert_called_once_with(
            region=Region.US, frequency=Frequency.MONTHLY
        )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from datetime import date, timedelta
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Iterable

from numpy import dtype

from f4ratk.data_reader import (
    CsvFileReader,
    fama_french_reader,
    fred_reader,
    yahoo_reader,
)


class TestFamaReader:
    def should_create_fama_reader_for_given_returns_data_name(self):
        result = fama_french_reader(returns_data='Developed_5_Factors_Daily')
        assert result.symbols == 'Developed_5_Factors_Daily'

    def should_create_fama_reader_with_cache(self):
        result = fama_french_reader(returns_data='')

        assert result.session.expire_after == timedelta(days=14)
        assert result.session.cache.name.endswith('requests')

    def should_query_full_data_range_by_default(self):
        result = fama_french_reader(returns_data='')

        assert result.start == date(1920, 1, 1)


class TestYahooReader:
    def should_create_yahoo_reader_for_given_ticker_symbol_and_date_range(self):
        result = yahoo_reader(
            ticker_symbol="V6IC.DE", start=date(2014, 1, 1), end=date(2019, 12, 31)
        )
        assert result.symbols == 'V6IC.DE'
        assert result.start == date(2014, 1, 1)
        assert result.end == date(2019, 12, 31)

    def should_create_yahoo_reader_with_cache(self):
        result = yahoo_reader(ticker_symbol=None, start=None, end=None)

        assert result.session.expire_after == timedelta(days=14)
        assert result.session.cache.name.endswith('requests')

    def should_query_full_data_range_by_default(self):
        result = yahoo_reader(ticker_symbol=None, start=None, end=None)

        assert result.start == date(1970, 1, 1)


class TestFredReader:
    def should_create_fred_reader_for_given_exchange_symbol_and_date_range(self):
        result = fred_reader(
            exchange_symbol='DEXUSEU', start=date(2014, 1, 1), end=date(2019, 12, 31)
        )
        assert result.symbols == 'DEXUSEU'
        assert result.start == date(2014, 1, 1)
        assert result.end == date(2019, 12, 31)

    def should_create_fred_reader_with_cache(self):
        result = fred_reader(exchange_symbol=None, start=None, end=None)

        assert result.session.expire_after == timedelta(days=14)
        assert result.session.cache.name.endswith('requests')

    def should_query_full_data_range_by_default(self):
        result = fred_reader(exchange_symbol=None, start=None, end=None)

        assert result.start == date(1970, 1, 1)


class TestCsvFileReader:
    def when_read_file_with_content(self, lines: Iterable[str]):
        pass

    def should_parse_isodate_and_returns_column(self):
        with NamedTemporaryFile(mode='w+', prefix='f4ratk_test_tmp_') as file:
            print(
                *('2020-11-18,"15.35"', '2020-11-17,14.33'),
                sep='\n',
                end='\n',
                file=file
            )

            file.flush()

            result = CsvFileReader(path=Path(file.name)).read()

        assert len(result) == 2
        assert result.index.dtype == dtype('datetime64[ns]')
        assert result.index.name == 'Dates'
        assert list(result.columns) == ['Returns']

        assert result['Returns']['2020-11-18'][0] == 15.35
        assert result['Returns']['2020-11-17'][0] == 14.33



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import typing
from dataclasses import dataclass
from enum import Enum, auto, unique

from click import group
from flask import Flask, Response, make_response
from flask.cli import FlaskGroup

# noinspection PyPackageRequirements
from marshmallow import Schema, ValidationError
from marshmallow.fields import List, Nested, Str
from werkzeug.exceptions import BadRequest, HTTPException, NotFound

from f4ratk.infrastructure import (
    configure_logging,
    instantiate_dependencies as main_dependencies,
)
from f4ratk.ticker.api import NoTickerData
from f4ratk.web.controller import web_blueprint
from f4ratk.web.infrastructure import (
    instantiate_dependencies as web_dependencies,
    mail_enabled,
    server_cache,
)
from f4ratk.web.mail.controller import mail_blueprint


@unique
class Error(Enum):
    GENERAL = auto()
    NO_DATA = auto()
    INVALID = auto()


@dataclass(frozen=True)
class Source:
    pointer: str


@dataclass(frozen=True)
class ErrorMessage:
    code: str
    status: str
    title: str = None
    detail: str = None
    source: Source = None


@dataclass(frozen=True)
class ErrorsMessage:
    errors: typing.List[ErrorMessage]


class SourceSchema(Schema):
    pointer = Str()


class ErrorSchema(Schema):
    code = Str(required=True)
    status = Str(required=True)
    title = Str()
    detail = Str()
    source = Nested(SourceSchema)


class ErrorsSchema(Schema):
    errors = List(Nested(ErrorSchema))


def create_app() -> Flask:
    configure_logging(verbose=False, server=True)
    main_dependencies()
    web_dependencies()

    app = Flask('f4ratk-server', static_folder=None)

    server_cache.init_app(app)

    @app.errorhandler(HTTPException)
    def error_500(exception: HTTPException) -> Response:
        return make_response(
            ErrorsSchema().dump(
                ErrorsMessage(
                    errors=[
                        ErrorMessage(
                            code=Error.GENERAL.name,
                            status=str(exception.code),
                            title=str("General Error"),
                            detail=str(exception.description),
                        )
                    ]
                )
            ),
            exception.code,
        )

    @app.errorhandler(NoTickerData)
    def error_symbol_unknown(exception: NoTickerData) -> Response:

        return make_response(
            ErrorsSchema().dump(
                ErrorsMessage(
                    errors=[
                        ErrorMessage(
                            code=Error.NO_DATA.name,
                            status=str(NotFound.code),
                            title="No data",
                            detail=str(exception),
                        )
                    ]
                )
            ),
            NotFound.code,
        )

    @app.errorhandler(ValidationError)
    def error_validation(exception: ValidationError) -> Response:
        return make_response(
            ErrorsSchema().dump(
                ErrorsMessage(
                    errors=[
                        ErrorMessage(
                            code=Error.INVALID.name,
                            status=str(BadRequest.code),
                            title="Invalid Attribute",
                            detail=f"{description}",
                            source=Source(pointer=f"/{attribute}"),
                        )
                        for attribute, message in exception.messages.items()
                        for description in message
                    ]
                )
            ),
            BadRequest.code,
        )

    with app.app_context():
        app.register_blueprint(web_blueprint)

        if mail_enabled():
            app.register_blueprint(mail_blueprint)
    return app


@group(cls=FlaskGroup, create_app=create_app)
def server_main():
    pass
